from xlrd import open_workbook, xldate_as_tuple
from datetime import date
from sqlalchemy import *
from functools import partial


def date_convert(datestring, datemode):
    """Create a date from an excel file cell that is formatted as a date
    cell."""

    datetuple = xldate_as_tuple(datestring, datemode)
    d = date(datetuple[0], datetuple[1], datetuple[2])
    return d


def generate_column(description, columnwidth=300, datemode=0):
    """Take a description of a column and return a column object

    column titles will be in the format of
    <name>.<type> or...
    <name>.<type>.unique
    """
    if description != "id":
        try:
            name, col_type, is_unique = description.split(".")
        except:
            name, col_type = description.split(".")
    #determine if unique
    if 'is_unique' in locals():
        #double check if unique is "unique" and not "gravy"
        if is_unique == "unique":
            is_unique = True
        else:
            is_unique = False
    else:
        is_unique = False

    #going through the rows of the database
    #you may need a function to convert
    #the data to a value sqlalchemy can accept
    #if it's bread and butter data for a varchar or int - sqlaclhemy does fine
    #so create a lambda that just passes the input to the output.
    func = lambda x:x
    if description == "id":
        column = Column('id', Integer, primary_key=True)
    elif col_type == "text":
        column = Column(name, Unicode(columnwidth), unique=is_unique)
    elif col_type == "int":
        column = Column(name, Integer)
    elif col_type == "float":
        column = Column(name, Float)
    elif col_type == "date":
        column = Column(name, Date)
        func = partial(date_convert,datemode=datemode)
    return column,func


def xlsx_to_sqlite(filename):
    #guardian clauses to defend the realm
    if not filename.endswith("xlsx"):
        raise ValueError("Pish posh! That's not a good xlsx filename.")
    fname, extension = filename.split(".")
    outfile_name = 'sqlite:///%s.db' % fname
    engine = create_engine(outfile_name, echo=True)
    metadata = MetaData()

    book = open_workbook(filename)
    #the function table stores the conversion functions needed for each
    #worksheet
    function_table = {}
    #go through each page in the workbook - create a table for each sheet
    for sheet, name in zip(book.sheets(), book.sheet_names()):
        #print the columns of the sheet
        values = []
        t = Table(name, metadata)
        #keep a tab on each function needed to convert the xl data to good
        #sqlalchemy data
        conversion_functions = []
        #generate a column for each column in the worksheet
        for col in range(sheet.ncols):
            column_description = sheet.cell(0, col).value
            #find the longest string in the column if it's a text column
            max_len = 10
            #need to determine the needed width if the column is text - go
            #through the whole column and then set the  column width
            #to 3x the max value
            #found
            if "text" in column_description:
                for row in range(sheet.nrows):
                    cell_len = len(sheet.cell(row, col).value)
                    max_len = cell_len if cell_len > max_len else max_len
            #create the column and conversion function from the
            #column description
            column, func = generate_column(column_description,
                                           columnwidth=3*max_len,
                                           datemode=book.datemode)
            t.append_column(column)
            conversion_functions.append(func)
        function_table[name] = conversion_functions
    #create the database
    metadata.create_all(engine)
    conn = engine.connect()
    #insert the data from the table
    for sheet, name in zip(book.sheets(), book.sheet_names()):
        t = Table(name, metadata)
        conversion_functions = function_table[name]
        for row in range(1, sheet.nrows):
            #insert the data - and convert each cell in the row according to
            #the required conversion function
            stmt = t.insert([func(sheet.cell(row, col).value)
                            for col, func
                            in zip(range(sheet.ncols), conversion_functions)])
            res = conn.execute(stmt)
    conn.close()
