from sqlalchemy import create_engine, MetaData, and_, or_
import sqlalchemy
from flask import Flask, render_template, url_for
from flask import session, request, abort, redirect
from os import path
import random
from functools import wraps
from datetime import date
import string
import requests
import json
from math import ceil
import locale
from columns import varchar_filter, integer_filter, float_filter, date_filter
from columns import add_varchar_filtered_columns
from columns import add_integer_filtered_columns
from columns import add_float_filtered_columns
from columns import add_date_filtered_columns
locale.setlocale(locale.LC_ALL, '')
PERPAGE = 5


def generate_csrf_token():
    """Generate a csrf token and put it in the session. Return the token"""

    session['_csrf_token'] = ''.join(random.choice(string.ascii_letters
                                     + string.punctuation
                                     + string.digits) for x in range(64))
    return session['_csrf_token']


def csrf_protect(f):
    """Wrap a function with a decorator that will return a 400 abort if the
    forms _csrf token doesn't match the csrf token"""

    @wraps(f)
    def csrf_protected(*args, **kwargs):
        if request.method == "POST":
            try:
                val =  session.pop("_csrf_token")
                if(val != request.form['_csrf_token']):
                    abort(400)
            except:
                abort(400)
        return f(*args, **kwargs)
    return csrf_protected


def login_required(f):
    """Decorator that requires the user to be logged in."""

    @wraps(f)
    def login_req(*args, **kwargs):
        if not session.get('email') or not session.get('logged_in'):
            abort(401)
        return f(*args, **kwargs)
    return login_req


def table_view(engine, table, table_name):
    """Return a function to view a table.

    Taking a db engine, the table, and table_name - generate a view for the
    user.

    The view will have UPDATE,READ, and DELETE functions attached to reach
    row.  There will be some basic pagination."""

    global PERPAGE

    @login_required
    def view(page):
        conn = engine.connect()
        stmt = table.select().count()
        num_items = conn.execute(stmt).fetchone()[0]
        pages = int(ceil(num_items / float(PERPAGE)))
        stmt = table.select(group_by=table.c.id, offset=(page-1)*PERPAGE)
        res = conn.execute(stmt).fetchmany(PERPAGE)
        return render_template("table.html",
                               num_items=num_items,
                               table_name=table_name,
                               results=res,
                               table=table,
                               page=page,
                               pages=pages)
    return view


def table_filter(engine, table, table_name):
    """Generate a function that provides a page to generate filters
       and view the results.

    Same as the table_view function but go through the table and then
    pick out the form's arguments to run a cumulative sqlalchemy filter.
    Execute once to count, then again to get the results.
    There will be basic pagination.
    """
    global PERPAGE

    @login_required
    def filter(page):
        conn = engine.connect()
        where_items = []
        if request.args:
            args = request.args
        else:
            return render_template("filter.html",
                                   table_name=table_name,
                                   table=table,
                                   page=page)
        filtered_columns = []
        for col in table.columns:
            #col_type= getattr(table.c,arg).type
            col_type = col.type
            this_clause = None
            if "VARCHAR" in str(col_type):
                filtered_columns = add_varchar_filtered_columns(filtered_columns, col, args)
                this_clause = varchar_filter(table, col, args)
                if this_clause is None: continue
            elif "INTEGER" in str(col_type):
                filtered_columns = add_integer_filtered_columns(filtered_columns, col, args)
                this_clause = integer_filter(table,col,args)
            elif "FLOAT" in str(col_type):
                filtered_columns = add_float_filtered_columns(filtered_columns, col, args)
                this_clause = float_filter(table,col, args)
            elif "DATE" in str(col_type):
                filtered_columns = add_date_filtered_columns(filtered_columns, col, args)
                this_clause = date_filter(table,col, args)
            where_items.append(this_clause)

        whereclause = and_(*where_items)
        query = table.select(whereclause).count()
        num_items = conn.execute(query).fetchone()[0]
        pages = int(ceil(num_items / float(PERPAGE)))
        query = table.select(whereclause, limit=PERPAGE, offset=(page-1)*PERPAGE)
        res = conn.execute(query).fetchall()
        return render_template("filter.html",
                               table_name=table_name,
                               results=res,
                               table=table,
                               page=page,
                               pages=pages,
                               num_items=num_items,
                               args=args,
                               filtered_columns=filtered_columns)
    return filter


def verify_form(table, request):
    """Roll over a table and it's request and return a list of errors and the
     proper values that can go straight into an sqlalchemy table.

    This is primarily used in item_create - but can be used elsewhere.
    """

    errors = {}
    #verify that the data is good
    cols = table.columns
    values = {}
    for col in cols:
        if col.name == "id":
            #let sqlalchemy put in the id
            continue
        try:
            if len(request.form[col.name]) == 0:
                raise ValueError
            #test the request based off the type
            if type(col.type) == sqlalchemy.types.DATE:
                month, day, year = request.form[col.name].split("/")
                d = date(int(year), int(month), int(day))
                values[col.name] = d
            elif type(col.type) == sqlalchemy.types.VARCHAR:
                values[col.name] = str(request.form[col.name])
            elif type(col.type) == sqlalchemy.types.INTEGER:
                if col.name == "price":
                    val = float(request.form[col.name]) * 100
                    values[col.name] = int(val)
                else:
                    values[col.name] = int(request.form[col.name])
            elif type(col.type) == sqlalchemy.types.FLOAT:
                values[col.name] = float(request.form[col.name])
            else:
                values[col.name] = request.form[col.name]
        except ValueError:
            errors[col.name] = ("I don't know how to work with %s going into %s."
                                % (request.form[col.name], col.name))
    return errors, values


def item_create(engine, table, table_name):
    """Return function that would view a form that would create an item
    and.  If the function receives as post - then verify and create the item
    based on the form data - redirect to the table view."""
    @login_required
    @csrf_protect
    def form_view(mod_id=None):
        if request.method == "GET" and (mod_id is None):
            for col in table.c:
                print(type(col.type))
        if request.method == "GET" and (mod_id is not None):
            #get the data for the item
            conn = engine.connect()
            stmt = table.select().where(table.c.id == mod_id)
            mod_data = conn.execute(stmt).fetchone()
            conn.close()
            return render_template("create.html",
                                   table_name=table_name,
                                   table=table,
                                   modify_item=True,
                                   mod_data=mod_data,
                                   mod_id=mod_id)
        elif request.method == "POST" and (mod_id is None):
            errors, values = verify_form(table, request)
            if len(errors.keys()) > 0:
                return render_template("create.html",
                                       table_name=table_name,
                                       table=table,
                                       errors=errors,
                                       vals=request.form)
            else:
                conn = engine.connect()
                conn.execute(table.insert(), [values])
                conn.close()
                return redirect(url_for(str("%s_view" % (table_name))))
        elif request.method == "POST" and (mod_id is not None):
            errors, values = verify_form(table, request)
            conn = engine.connect()
            try:
                conn.execute(table.update().where(table.c.id == request.form['id']).values(values))
            except sqlalchemy.exc.IntegrityError:
                conn.close()
                return "You tried to update an item - but that name was somewhere else in the database."
            except:
                conn.close()
                return "Failed to update."
            else:
                conn.close()
                return redirect(url_for(str("%s_view" % (table_name))))
        return render_template("create.html",
                               table_name=table_name,
                               table=table)
    return form_view


def item_view(engine, table, table_name, delete_command=False):
    """Return a function that would view an individual item in a page."""
    @login_required
    @csrf_protect
    def view(id_num):
        if request.method == "GET":
            conn = engine.connect()
            stmt = table.select().where(table.c.id == id_num)
            res = conn.execute(stmt).fetchone()
            conn.close()
            if res:
                return render_template("item.html",
                                       table_name=table_name,
                                       id_num=id_num,
                                       results=res,
                                       table=table,
                                       delete_command=delete_command)
            else:
                return render_template("item.html",
                                       table_name=table_name,
                                       id_num=id_num,
                                       table=table)
        if request.method == "POST" and request.form['delete'] == 'delete_item':
            id_del = int(request.form['id'])
            conn = engine.connect()
            stmt = table.delete().where(table.c.id == id_del)
            try:
                res = conn.execute(stmt)
                conn.close()
            except:
                return "Looks like you want to delete item %i from %s, but I couldn't"
            else:
                return redirect(url_for(str("%s_view" % (table_name))))
    return view


def present(constring, AUDIENCE):
    """Create an engine from a sqlite db file and explore it's metadata to
    create a usuable basic flask app with authorisation and CRUD capabilities.

    The AUDIENCE file is the website that the site will be served on - it's
    needed by persona.
    """
    engine = create_engine(constring, echo=False)
    metadata = MetaData()
    metadata.reflect(bind=engine)
    #create the flask app with the templates local to this code
    app = Flask(__name__,
                template_folder=path.join(path.dirname(__file__), "templates"),
                static_folder=path.join(path.dirname(__file__), "static"))
    #add items to the jinja globals so the templates can be a little smarter.
    app.jinja_env.globals.update(type=type)
    app.jinja_env.globals.update(int=int)
    app.jinja_env.globals.update(float=float)
    app.jinja_env.globals.update(sqlalchemy=sqlalchemy)
    app.jinja_env.globals.update(getattr=getattr)
    app.jinja_env.globals.update(locale=locale)
    app.jinja_env.globals.update(generate_csrf_token=generate_csrf_token)
    #create pages for each table
    table_refs = []
    #for each table in db create CRUD pages
    for tab_name in metadata.tables:
        table = metadata.tables[tab_name]
        #view paths
        app.add_url_rule("/%s_view_page/<int:page>" % tab_name,
                         "%s_view_page" % tab_name,
                         table_view(engine, table, tab_name))
        app.add_url_rule("/%s_view" % tab_name,
                         "%s_view" % tab_name,
                         table_view(engine, table, tab_name),
                         defaults={'page': 1})
        app.add_url_rule("/%s_view_id/<int:id_num>" % tab_name,
                         "%s_view_id" % tab_name,
                         item_view(engine, table, tab_name))
        #filter path
        app.add_url_rule("/%s_filter" % tab_name,
                         "%s_filter" % tab_name,
                         table_filter(engine, table, tab_name),
                         defaults={'page': 1})
        app.add_url_rule("/%s_filter_results/<int:page>" % tab_name,
                         "%s_filter_results" % tab_name,
                         table_filter(engine, table, tab_name))
        #delete path
        app.add_url_rule("/%s_delete_id/<int:id_num>" % tab_name,
                         "%s_delete_id" % tab_name,
                         item_view(engine, table, tab_name, delete_command=True),
                         methods=['GET', 'POST'])
        create_mod_form_view = item_create(engine, table, tab_name)
        #create path
        app.add_url_rule("/%s_create" % tab_name,
                         "%s_create" % tab_name,
                         create_mod_form_view,
                         methods=['GET', 'POST'])
        #mod path
        app.add_url_rule("/%s_modify_id/<int:mod_id>" % tab_name,
                         "%s_modify" % tab_name,
                         create_mod_form_view,
                         methods=['GET', 'POST'])
        table_refs.append({"name": tab_name,
                          "endpoint": "%s_view" % tab_name})

    #create the non-table pages
    #create the index page
    def index():
        return render_template("main.html",
                               tables=table_refs,
                               db_name=constring.split("/")[-1])
    app.add_url_rule("/", "index", index)

    #add the login pages

    def login():
        if 'assertion' not in request.form:
            abort(400)
        data = {'assertion': request.form['assertion'], 'audience': AUDIENCE}
        res = requests.post('https://verifier.login.persona.org/verify', data=data)
        if res.ok:
            verification = json.loads(res.content)
            if verification['status'] == 'okay':
                #query the user table and see if the email is in the db
                session.update({'email': verification['email']})
                if 'users' in metadata.tables:
                    table = metadata.tables['users']
                    conn = engine.connect()
                    stmt = table.select().where(table.c.email == verification['email'])
                    email_res = conn.execute(stmt).fetchone()
                    conn.close()
                    if email_res:
                        session.update({'logged_in': True})
                else:
                    session.update({'logged_in': True})

                return res.content
        abort(500)
    app.add_url_rule("/auth/login", "login", login, methods=['POST'])

    def logout():
        session.pop('email', None)
        session.pop('logged_in', None)
        return redirect('/')
    app.add_url_rule("/auth/logout", "logout", logout, methods=['POST'])

    return app
